import vertexai, os, traceback, argparse
from vertexai.preview.generative_models import GenerativeModel
from vertexai.generative_models._generative_models import (
    GenerationConfig,
    HarmCategory,
    HarmBlockThreshold,
)
from searchbible import config
#from searchbible.utils.bible_studies import bible_study_suggestions
from searchbible.utils.streaming_word_wrapper import StreamingWordWrapper
from searchbible.health_check import HealthCheck
if not hasattr(config, "exit_entry"):
    HealthCheck.setBasicConfig()
    HealthCheck.saveConfig()
    print("Updated!")
#HealthCheck.setPrint()
from prompt_toolkit.styles import Style
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.shortcuts import clear
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.key_binding import KeyBindings
from pathlib import Path
import threading


# Install google-cloud-aiplatform FIRST!
#!pip install --upgrade google-cloud-aiplatform


class GeminiPro:

    def __init__(self, name="Gemini Pro", temperature=0.9, max_output_tokens=8192):
        # authentication
        authpath1 = os.path.join(HealthCheck.getFiles(), "credentials_googleaistudio.json")
        if os.path.isfile(authpath1):
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = authpath1
            self.runnable = True
        else:
            print(f"API key json file '{authpath1}' not found!")
            print("Read https://github.com/eliranwong/letmedoit/wiki/Google-API-Setup for setting up Google API.")
            self.runnable = False
        # initiation
        vertexai.init()
        self.name, self.temperature = name, temperature
        self.generation_config=GenerationConfig(
            temperature=temperature, # 0.0-1.0; default 0.9
            max_output_tokens=max_output_tokens, # default
            candidate_count=1,
        )
        self.safety_settings={
            HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: HarmBlockThreshold.BLOCK_NONE,
            HarmCategory.HARM_CATEGORY_HATE_SPEECH: HarmBlockThreshold.BLOCK_NONE,
            HarmCategory.HARM_CATEGORY_HARASSMENT: HarmBlockThreshold.BLOCK_NONE,
            HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT: HarmBlockThreshold.BLOCK_NONE,
        }
        self.defaultPrompt = ""
        #self.enableVision = (os.path.realpath(__file__).endswith("vision.py"))

    def run(self, prompt=""):
        completer = WordCompleter(
            config.read_suggestions + ["[bible]"],
            ignore_case=True,
            sentence=True,
        )
        custom_key_bindings = KeyBindings()

        @custom_key_bindings.add("c-n")
        def _(event):
            buffer = event.app.current_buffer
            buffer.text = ".new"
            buffer.validate_and_handle()

        if self.defaultPrompt:
            prompt, self.defaultPrompt = self.defaultPrompt, ""
        historyFolder = os.path.join(HealthCheck.getFiles(), "history")
        Path(historyFolder).mkdir(parents=True, exist_ok=True)
        chat_history = os.path.join(historyFolder, "geminipro")
        chat_session = PromptSession(history=FileHistory(chat_history))

        promptStyle = Style.from_dict({
            # User input (default text).
            "": config.terminalCommandEntryColor2,
            # Prompt.
            "indicator": config.terminalPromptIndicatorColor2,
        })

        #completer = WordCompleter(["[", "[NO_FUNCTION_CALL]"], ignore_case=True) if self.enableVision else None

        if not self.runnable:
            print(f"{self.name} is not running due to missing configurations!")
            return ""
        model = GenerativeModel("gemini-pro")
        chat = model.start_chat()
        justStarted = True
        HealthCheck.print2(f"\n{self.name} loaded!")
        print("(To start a new chart, enter '.new')")
        print(f"(To quit, enter '{config.exit_entry}')\n")
        while True:
            if not prompt:
                prompt = HealthCheck.simplePrompt(
                    style=promptStyle,
                    promptSession=chat_session,
                    completer=completer,
                    custom_key_bindings=custom_key_bindings,
                    bottom_toolbar=" [ctrl+q] exit [ctrl+n] new chat ",
                )
            else:
                prompt = HealthCheck.simplePrompt(
                    style=promptStyle,
                    promptSession=chat_session,
                    completer=completer,
                    custom_key_bindings=custom_key_bindings,
                    bottom_toolbar=" [ctrl+q] exit [ctrl+n] new chat ",
                    default=prompt,
                    accept_default=True,
                )
            if prompt == config.exit_entry:
                break
            elif prompt == ".new":
                clear()
                chat = model.start_chat()
                justStarted = True
                print("New chat started!")
            elif "[bible]" in prompt:
                HealthCheck.print2(f"\n{self.name} closed!")
                return prompt.replace("[bible]", "").strip()
            elif prompt := prompt.strip():
                prompt = config.removeSpecialEntries(prompt)
                if prompt and not prompt in (".new", config.exit_entry) and hasattr(config, "currentMessages"):
                    userMessage = {"role": "user", "content": prompt}
                    config.currentMessages.append(userMessage)

                streamingWordWrapper = StreamingWordWrapper()
                config.pagerContent = ""
                #self.addPagerContent = True

                # declare a function
#                get_vision_func = generative_models.FunctionDeclaration(
#                    name="analyze_images",
#                    description="Describe or analyze images. Remember, do not use this function for non-image related tasks. Even it is an image-related task, use this function ONLY if I provide at least one image file path or image url.",
#                    parameters={
#                        "type": "object",
#                        "properties": {
#                            "query": {
#                                "type": "string",
#                                "description": "Questions or requests that users ask about the given images",
#                            },
#                            "files": {
#                                "type": "string",
#                                "description": """Return a list of image paths or urls, e.g. '["image1.png", "/tmp/image2.png", "https://letmedoit.ai/image.png"]'. Return '[]' if image path is not provided.""",
#                            },
#                        },
#                        "required": ["query", "files"],
#                    },
#                )
#                vision_tool = generative_models.Tool(
#                    function_declarations=[get_vision_func],
#                )

                try:
                    if justStarted:
                        prompt = f"{config.aiSystemMessage}\n{prompt}"
                        justStarted = False
                    # https://cloud.google.com/vertex-ai/docs/generative-ai/model-reference/gemini
                    # Note: At the time of writing, function call feature with Gemini Pro is very weak, compared with the function call feature offerred by ChatGPT:
                    # 1. Gemini Pro do not accept multiple tools in a single message
                    # 2. Gemini Pro is weak to determine if it is appropriate to use the given tool or not.  When a tool is given, it is called by mistake so often.  In contrast, ChatGPT has the "auto" setting which makes ChatGPT obviously smarter than Gemini Pro.
                    #if "[NO_FUNCTION_CALL]" in prompt or not self.enableVision:
                    #    allow_function_call = False
                    #    prompt = prompt.replace("[NO_FUNCTION_CALL]", "")
                    #else:
                    #    allow_function_call = True
                    completion = chat.send_message(
                        prompt,
                        # Optional:
                        generation_config=self.generation_config,
                        safety_settings=self.safety_settings,
                        #tools=[vision_tool] if allow_function_call else None,
                        stream=True,
                    )

                    # Create a new thread for the streaming task
                    streaming_event = threading.Event()
                    self.streaming_thread = threading.Thread(target=streamingWordWrapper.streamOutputs, args=(streaming_event, completion,))
                    # Start the streaming thread
                    self.streaming_thread.start()

                    # wait while text output is steaming; capture key combo 'ctrl+q' or 'ctrl+z' to stop the streaming
                    streamingWordWrapper.keyToStopStreaming(streaming_event)

                    # when streaming is done or when user press "ctrl+q"
                    self.streaming_thread.join()

                    # format response when streaming is not applied
                    #tokens = list(pygments.lex(fullContent, lexer=MarkdownLexer()))
                    #print_formatted_text(PygmentsTokens(tokens), style=HealthCheck.getPygmentsStyle())

                except:
                    self.streaming_thread.join()
                    HealthCheck.print2(traceback.format_exc())

            prompt = ""

        HealthCheck.print2(f"\n{self.name} closed!")
        return ""

def main():
    # Create the parser
    parser = argparse.ArgumentParser(description="geminipro cli options")
    # Add arguments
    parser.add_argument("default", nargs="?", default=None, help="default entry")
    parser.add_argument('-o', '--outputtokens', action='store', dest='outputtokens', help="specify maximum output tokens with -o flag; default: 8192")
    parser.add_argument('-t', '--temperature', action='store', dest='temperature', help="specify temperature with -t flag: default: 0.9")
    # Parse arguments
    args = parser.parse_args()
    # Get options
    prompt = args.default.strip() if args.default and args.default.strip() else ""
    if args.outputtokens and args.outputtokens.strip():
        try:
            max_output_tokens = int(args.outputtokens.strip())
        except:
            max_output_tokens = 8192
    else:
        max_output_tokens = 8192
    if args.temperature and args.temperature.strip():
        try:
            temperature = float(args.temperature.strip())
        except:
            temperature = 0.9
    else:
        temperature = 0.9
    GeminiPro(
        temperature=temperature,
        max_output_tokens = max_output_tokens,
    ).run(
        prompt=prompt,
    )

if __name__ == '__main__':
    main()


from searchbible import config
import pprint, os, shutil
from prompt_toolkit.shortcuts import yes_no_dialog

def setConfig(defaultSettings, thisTranslation={}, temporary=False):
    for key, value in defaultSettings:
        if not hasattr(config, key):
            value = pprint.pformat(value)
            exec(f"""config.{key} = {value} """)
            if temporary:
                config.excludeConfigList.append(key)
    if thisTranslation:
        for i in thisTranslation:
            if not i in config.thisTranslation:
                config.thisTranslation[i] = thisTranslation[i]

defaultSettings = (
    ('aiSystemMessage', "You're an expert on the Bible.\nPlease respond to all my inquiries with spiritual insights and informed knowledge regarding biblical teachings.\nPlease always strive to maintain an engaging and continuous conversation.\n"),
    ('thisTranslation', {}),
    ('cancel_entry', '.cancel'),
    ('exit_entry', '.exit'),
    ('terminalHeadingTextColor', 'ansigreen'),
    ('terminalResourceLinkColor', 'ansiyellow'),
    ('terminalCommandEntryColor1', 'ansiyellow'),
    ('terminalPromptIndicatorColor1', 'ansimagenta'),
    ('terminalCommandEntryColor2', 'ansigreen'),
    ('terminalPromptIndicatorColor2', 'ansicyan'),
    ('terminalSearchHighlightBackground', 'ansiblue'),
    ('terminalSearchHighlightForeground', 'ansidefault'),
    ('embeddingModel', 'paraphrase-multilingual-mpnet-base-v2'),
    ('customTextEditor', ""), # e.g. 'micro -softwrap true -wordwrap true'; built-in text editor eTextEdit is used when it is not defined.
    ('pagerView', False),
    #('usePygame', True),
    ('wrapWords', True),
    ('mouseSupport', False),
    ('autoUpgrade', True),
    ('chatGPTApiModel', 'gpt-3.5-turbo-16k'),
    ('chatGPTApiMaxTokens', 2000),
    ('chatGPTApiMinTokens', 256),
    ('llmTemperature', 0.8),
    ('openaiApiKey', ''),
    ('openaiApiOrganization', ''),
    ('predefinedContext', '[none]'),
    ('customPredefinedContext', ''),
    ('applyPredefinedContextAlways', False), # True: apply predefined context with all use inputs; False: apply predefined context only in the beginning of the conversation
    ('pygments_style', ''),
    ('developer', False),
 
    ("mainText", "NET"),
    ("mainB", 43),
    ("mainC", 3),
    ("mainV", 16),
    ("enableCaseSensitiveSearch", False),
    ("noWordWrapBibles", []), # some bibles display better, without word wrap feature, e.g. CUV
    ("chapterParagraphsAndSubheadings", True),

    ('terminalEnableTermuxAPI', False),
    ('terminalEditorScrollLineCount', 20),
    ('terminalEditorTabText', "    "),
    ('blankEntryAction', "..."),
    ('storagedirectory', ""),
    ('vlcSpeed', 1.0),
    ('gcttsLang', "en-GB"),
    ('gcttsSpeed', 1.0),
    ('gttsLang', "en"), # gTTS is used by default if ttsCommand is not given
    ('gttsTld', ""), # https://gtts.readthedocs.io/en/latest/module.html#languages-gtts-lang
    ('ttsCommand', ""), # ttsCommand is used if it is given; offline tts engine runs faster; on macOS [suggested speak rate: 100-300], e.g. "say -r 200 -v Daniel"; on Ubuntu [espeak; speed in approximate words per minute; 175 by default], e.g. "espeak -s 175 -v en"; remarks: always place the voice option, if any, at the end
    ('ttsCommandSuffix', ""), # try on Windows; ttsComand = '''Add-Type -TypeDefinition 'using System.Speech.Synthesis; class TTS { static void Main(string[] args) { using (SpeechSynthesizer synth = new SpeechSynthesizer()) { synth.Speak(args[0]); } } }'; [TTS]::Main('''; ttsCommandSuffix = ")"; a full example is Add-Type -TypeDefinition 'using System.Speech.Synthesis; class TTS { static void Main(string[] args) { using (SpeechSynthesizer synth = new SpeechSynthesizer()) { synth.Speak(args[0]); } } }'; [TTS]::Main("Text to be read")
    ("ttsLanguages", ["en", "en-gb", "en-us", "zh", "yue", "el"]), # users can edit this item in config.py to support more or less languages
    ("ttsLanguagesCommandMap", {"en": "", "en-gb": "", "en-us": "", "zh": "", "yue": "", "el": "",}), # advanced users need to edit this item manually to support different voices with customised tts command, e.g. ttsCommand set to "say -r 200 -v Daniel" and ttsLanguagesCommandMap set to {"en": "Daniel", "en-gb": "Daniel", "en-us": "", "zh": "", "yue": "", "el": "",}

    ('noOfLinesPerChunkForParsing', 100),
    ('parseEnglishBooksOnly', False),
    ('useLiteVerseParsing', False),
    ('standardAbbreviation', "ENG"),
    ('convertChapterVerseDotSeparator', True),
    ('parseBookChapterWithoutSpace', True),
    ('parseBooklessReferences', True),
    ('parseClearSpecialCharacters', False),
    ('parserStandarisation', "NO"),

    ('terminalEditorScrollLineCount', 20),
    ('maxClosestMatches', 20),
    ('compareBibles', False),
    ('compareBibleVersions', ["KJV", "NET"]),

    ("chatbot", "chatgpt"),
)

# save default configs
configFile = os.path.join(config.packageFolder, "config.py")
backupFile = os.path.join(config.storagedirectory, "config_backup.py")
if os.path.getsize(configFile) == 0:
    if os.path.isfile(backupFile):
        restore_backup = yes_no_dialog(
            title="Configuration Backup Found",
            text=f"Do you want to use the following backup?\n{backupFile}"
        ).run()
        if restore_backup:
            shutil.copy(backupFile, configFile)
            print("Configuration backup restored!")
            config.restartApp()
# load default config
setConfig(defaultSettings)
# share setConfig, to allow plugins to add customised config
config.setConfig = setConfig
# back up newly saved config
if os.path.getsize(configFile) == 0:
    config.saveConfig()
    shutil.copy(configFile, backupFile)



from searchbible.health_check import HealthCheck
from searchbible.converter.bible import ConvertBible
from searchbible import config
import argparse, os

def main():
    # set up basic configs
    if not hasattr(config, "exit_entry"):
        HealthCheck.setBasicConfig()

    # Create the parser
    parser = argparse.ArgumentParser(description="SearchBibleAI Converter CLI options")
    # Add arguments
    #parser.add_argument("default", nargs="?", default=None, help="default entry")
    parser.add_argument('-b', '--bible', action='store', dest='bible', help="convert Unique Bible App *.bible database; accepts a file or folder path")
    # Parse arguments
    args = parser.parse_args()
    # Get options
    #prompt = args.default.strip() if args.default and args.default.strip() else ""

    if givenPath := args.bible.strip():
        if os.path.isfile(givenPath):
            ConvertBible.convert_bible(givenPath)
        elif os.path.isdir(givenPath):
            for i in os.listdir(givenPath):
                iFile = os.path.join(givenPath, i)
                if os.path.isfile(iFile) and iFile.endswith(".bible"):
                    ConvertBible.convert_bible(iFile)


if __name__ == '__main__':
    main()


class RefUtil:

    @staticmethod
    def getAllRefFilters(refs: list) -> dict:
        filters = []
        for ref in refs:
            filters += RefUtil.getRefFilter(ref)
        if filters and len(filters) == 1:
            return filters[0]
        elif filters and len(filters) > 1:
            return {"$or": filters}
        else:
            return {}

    @staticmethod
    def getRefFilter(ref: tuple) -> list:
        if len(ref) == 3:
            b, c, v = ref
            return [{"$and": [{"book": {"$eq": b}}, {"chapter": {"$eq": c}}, {"verse": {"$eq": v}}]}]
        elif len(ref) == 5:
            b, c, v, c2, v2 = ref
            if c == c2:
                return [{"$and": [{"book": {"$eq": b}}, {"chapter": {"$eq": c}}, {"verse": {"$gte": v}}, {"verse": {"$lte": v2}}]}]
            elif c2 > c:
                filter = []
                filter.append({"$and": [{"book": {"$eq": b}}, {"chapter": {"$eq": c}}, {"verse": {"$gte": v}}]})
                filter.append({"$and": [{"book": {"$eq": b}}, {"chapter": {"$eq": c2}}, {"verse": {"$lte": v2}}]})
                if (c2 - c) > 1:
                    for i in range(c+1, c2):
                        filter.append({"$and": [{"book": {"$eq": b}}, {"chapter": {"$eq": i}}]})
                return filter
        else:
            return []


from searchbible.health_check import HealthCheck
from searchbible.converter.bible import ConvertBible
from searchbible import config
from searchbible.utils.RefUtil import RefUtil
from chromadb.config import Settings
import chromadb, os
from packaging import version


class Bible:

    @staticmethod
    def getBibleDir() -> str:
        ubaUserDir = os.path.join(os.path.expanduser("~"), "UniqueBible", "marvelData", "bibles")
        return ubaUserDir if os.path.isdir(ubaUserDir) else os.path.join(config.packageFolder, "data", "bibles")

    @staticmethod
    def getBibleList() -> list:
        bibles = os.path.join(config.storagedirectory, "bibles")
        return [] if not os.path.isdir(bibles) else [i for i in os.listdir(bibles) if os.path.isdir(os.path.join(bibles, i)) and os.path.isfile(os.path.join(bibles, i, "chroma.sqlite3"))]

    @staticmethod
    def getUbaBibleList() -> list:
        bibleDir = Bible.getBibleDir()
        return [i[:-6] for i in os.listdir(bibleDir) if os.path.isfile(os.path.join(bibleDir, i)) and i.endswith(".bible")]

    @staticmethod
    def getDbPath(bible: str) -> str:
        def convertBible(bible: str):
            HealthCheck.print3(f"Converting bible: {bible} ...")
            ConvertBible.convert_bible(os.path.join(Bible.getBibleDir(), f"{bible}.bible"))
        installedBibles = Bible.getBibleList()
        dbpath = os.path.join(config.storagedirectory, "bibles", bible)
        if bible in installedBibles:
            config.mainText = bible
            return dbpath
        elif bible in Bible.getUbaBibleList():
            convertBible(bible)
            config.mainText = bible
            return dbpath
        else:
            bible = "NET"
            if not bible in installedBibles:
                convertBible(bible)
            config.mainText = "NET"
            return os.path.join(config.storagedirectory, "bibles", "NET")

    @staticmethod
    def getSingleItem(ref: str, bible: str, paragraphs: bool=False) -> tuple:
        # get either a single verse or a single section of paragraphs
        # expect ref in the format, like "1.1.1"
        # keep mainText
        mainText = config.mainText

        dbpath = Bible.getDbPath(bible)
        if not dbpath:
            config.mainText = mainText
            return ()
        chroma_client = chromadb.PersistentClient(dbpath, Settings(anonymized_telemetry=False))
        collection = chroma_client.get_or_create_collection(
            name="paragraphs" if paragraphs else "verses",
            metadata={"hnsw:space": "cosine"},
            embedding_function=HealthCheck.getEmbeddingFunction(embeddingModel="paraphrase-multilingual-mpnet-base-v2"),
        )
        try:
            if paragraphs:
                res = collection.get(where={"start": ref})
                metadata = res["metadatas"][0]
                document = res["documents"][0]
                config.mainText = mainText
                return (metadata["start"], metadata["book_start"], metadata["chapter_start"], metadata["verse_start"], metadata["chapter_end"], metadata["verse_end"], document)
            else:
                res = collection.get(where={"reference": ref})
                metadata = res["metadatas"][0]
                document = res["documents"][0]
                config.mainText = mainText
                return (metadata["reference"], metadata["book"], metadata["chapter"], metadata["verse"], document)
        except:
            config.mainText = mainText
            return ()   

    @staticmethod
    def getVerses(refs: list, bible: str = "NET") -> list:
        isChapter = (len(refs) == 1 and len(refs[0]) == 3)

        filters = RefUtil.getAllRefFilters(refs)
        if not filters:
            return []
        #dbpath
        dbpath = Bible.getDbPath(bible)
        if not dbpath:
            return []
        # client
        chroma_client = chromadb.PersistentClient(dbpath, Settings(anonymized_telemetry=False))
        # collection
        collection = chroma_client.get_or_create_collection(
            name="verses",
            metadata={"hnsw:space": "cosine"},
            embedding_function=HealthCheck.getEmbeddingFunction(embeddingModel="paraphrase-multilingual-mpnet-base-v2"),
        )
        res = collection.get(where=filters)
        # unpack data
        metadatas = res["metadatas"]
        documents = res["documents"]
        verses = [(metadata["reference"], metadata["book"], metadata["chapter"], metadata["verse"], document) for metadata, document in zip(metadatas, documents)]
        # sorting
        verses = sorted(verses, key=lambda x: version.parse(x[0]))

        # check if it is a single reference; get also all verses in the chapter
        if isChapter:
            config.mainB, config.mainC, config.mainV = refs[0]
            res0 = collection.get(where={"$and": [{"book": {"$eq": config.mainB}}, {"chapter": {"$eq": config.mainC}}]})
            metadatas = res0["metadatas"]
            documents = res0["documents"]
            chapter = [(metadata["reference"], metadata["book"], metadata["chapter"], metadata["verse"], document) for metadata, document in zip(metadatas, documents)]
            return [sorted(chapter, key=lambda x: version.parse(x[0])), verses]
        else:
            return verses


terminalPromptIndicatorColor1 = "ansimagenta"
terminalPromptIndicatorColor2 = "ansicyan"
terminalCommandEntryColor1 = "ansiyellow"
terminalCommandEntryColor2 = "ansigreen"
terminalResourceLinkColor = "ansiyellow"
terminalHeadingTextColor = "ansigreen"


import apsw, uuid, os, chromadb, shutil
from chromadb.config import Settings
from searchbible import config
from searchbible.utils.AGBsubheadings import agbSubheadings
from searchbible.utils.AGBparagraphs_expanded import agbParagraphs
from searchbible.health_check import HealthCheck
from pathlib import Path
from prompt_toolkit.shortcuts import ProgressBar


class ConvertBible:
    """
    Convert UniqueBible App bible files (*.bible) into Chromadb format
    """

    @staticmethod
    def convert_bible(database: str) -> None:
        if not os.path.isfile(database) or not database.endswith(".bible"):
            print("Invalid file path given!")
            return None
        else:
            version = os.path.basename(database)[:-6]
        
        def getAllVerses():
            # get all verses from Unique Bible App bible file
            with apsw.Connection(database) as connection:
                cursor = connection.cursor()
                query = "SELECT * FROM Verses ORDER BY Book, Chapter, Verse"
                cursor.execute(query)
                allVerses = cursor.fetchall()
                return allVerses
            return ()

        # database path
        dbpath = os.path.join(config.storagedirectory, "bibles", version)
        if os.path.isdir(dbpath):
            # remove old database if it exists
            HealthCheck.print3(f"Removing old database: {dbpath}")
            shutil.rmtree(dbpath, ignore_errors=True)
        HealthCheck.print3(f"Creating database: {dbpath}")
        # create directory for chromadb files
        Path(dbpath).mkdir(parents=True, exist_ok=True)
        # keep all original data
        shutil.copy(database, os.path.join(dbpath, "chroma.sqlite3"))
        # client
        chroma_client = chromadb.PersistentClient(dbpath, Settings(anonymized_telemetry=False))
        # collection
        collectionVerse = chroma_client.get_or_create_collection(
            name="verses",
            metadata={"hnsw:space": "cosine"},
            embedding_function=HealthCheck.getEmbeddingFunction(embeddingModel="paraphrase-multilingual-mpnet-base-v2"),
        )
        collectionParagraph = chroma_client.get_or_create_collection(
            name="paragraphs",
            metadata={"hnsw:space": "cosine"},
            embedding_function=HealthCheck.getEmbeddingFunction(embeddingModel="paraphrase-multilingual-mpnet-base-v2"),
        )

        paragraphTitle = ""
        paragraphStart = ""
        paragraphStartB = ""
        paragraphStartC = ""
        paragraphStartV = ""
        paragraphEnd = ""
        paragraphEndB = ""
        paragraphEndC = ""
        paragraphEndV = ""
        paragraphContent = ""            

        with ProgressBar() as pb:
            for book, chapter, verse, scripture in pb(getAllVerses()):
                bcv = f"{book}.{chapter}.{verse}"

                metadata = {
                    "book": book,
                    "chapter": chapter,
                    "verse": verse,
                    "reference": bcv,
                }
                id = str(uuid.uuid4())
                collectionVerse.add(
                    documents = [scripture],
                    metadatas = [metadata],
                    ids = [id]
                )

                if bcv in agbSubheadings:
                    if paragraphStart and paragraphEnd:
                        # save previous paragraph
                        metadata = {
                            "title": paragraphTitle,
                            "start": paragraphStart,
                            "book_start": paragraphStartB,
                            "chapter_start": paragraphStartC,
                            "verse_start": paragraphStartV,
                            "end": paragraphEnd,
                            "book_end": paragraphEndB,
                            "chapter_end": paragraphEndC,
                            "verse_end": paragraphEndV,
                        }
                        id = str(uuid.uuid4())
                        collectionParagraph.add(
                            documents = [paragraphContent],
                            metadatas = [metadata],
                            ids = [id]
                        )
                    paragraphTitle = agbSubheadings.get(bcv)
                    paragraphStart = bcv
                    paragraphStartB = book
                    paragraphStartC = chapter
                    paragraphStartV = verse
                    paragraphContent = f"{paragraphTitle}\n{chapter}:{verse} {scripture}"
                else:
                    if (book, chapter, verse) in agbParagraphs:
                        paragraphContent += "\n"
                    paragraphContent += f"\n{chapter}:{verse} {scripture}"
                paragraphEnd = bcv
                paragraphEndB = book
                paragraphEndC = chapter
                paragraphEndV = verse

            # save the last paragraph
            metadata = {
                "title": paragraphTitle,
                "start": paragraphStart,
                "book_start": paragraphStartB,
                "chapter_start": paragraphStartC,
                "verse_start": paragraphStartV,
                "end": paragraphEnd,
                "book_end": paragraphEndB,
                "chapter_end": paragraphEndC,
                "verse_end": paragraphEndV,
            }
            id = str(uuid.uuid4())
            collectionParagraph.add(
                documents = [paragraphContent],
                metadatas = [metadata],
                ids = [id]
            )

        return None



from searchbible import config
from searchbible.utils.prompt_validator import NumberValidator
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.styles import Style
from prompt_toolkit.shortcuts import input_dialog, radiolist_dialog, checkboxlist_dialog, message_dialog
from prompt_toolkit.completion import WordCompleter


class TerminalModeDialogs:

    def __init__(self, parent) -> None:
        self.parent = parent
        self.style = Style.from_dict(
            {
                "dialog": "bg:ansiblack",
                "dialog text-area": f"bg:ansiblack {config.terminalCommandEntryColor2}",
                "dialog text-area.prompt": config.terminalPromptIndicatorColor2,
                "dialog radio-checked": config.terminalResourceLinkColor,
                "dialog checkbox-checked": config.terminalResourceLinkColor,
                "dialog button.arrow": config.terminalResourceLinkColor,
                "dialog button.focused": f"bg:{config.terminalResourceLinkColor} ansiblack",
                "dialog frame.border": config.terminalResourceLinkColor,
                "dialog frame.label": f"bg:ansiblack {config.terminalResourceLinkColor}",
                "dialog.body": "bg:ansiblack ansiwhite",
                "dialog shadow": "bg:ansiblack",
            }
        ) if config.terminalResourceLinkColor.startswith("ansibright") else Style.from_dict(
            {
                "dialog": "bg:ansiwhite",
                "dialog text-area": f"bg:ansiblack {config.terminalCommandEntryColor2}",
                "dialog text-area.prompt": config.terminalPromptIndicatorColor2,
                "dialog radio-checked": config.terminalResourceLinkColor,
                "dialog checkbox-checked": config.terminalResourceLinkColor,
                "dialog button.arrow": config.terminalResourceLinkColor,
                "dialog button.focused": f"bg:{config.terminalResourceLinkColor} ansiblack",
                "dialog frame.border": config.terminalResourceLinkColor,
                "dialog frame.label": f"bg:ansiwhite {config.terminalResourceLinkColor}",
                "dialog.body": "bg:ansiwhite ansiblack",
                "dialog shadow": "bg:ansiwhite",
            }
        )

    # a wrapper to standard input_dialog; open radiolist_dialog showing available options when user input is not a valid option.
    def searchableInput(self, title="Text Entry", text="Enter / Search:", default="", completer=None, options=[], descriptions=[], validator=None, numberOnly=False, password=False, ok_text="OK", cancel_text="Cancel"):
        if completer is None and options:
            completer = WordCompleter(options, ignore_case=True)
        if validator is None and numberOnly:
            validator=NumberValidator()
        result = input_dialog(
            title=title,
            text=text,
            default=default,
            completer=completer,
            validator=validator,
            password=password,
            ok_text=ok_text,
            cancel_text=cancel_text,
            style=self.style,
        ).run().strip()
        if result.lower() == config.cancel_entry:
            return result
        if options:
            if result and result in options:
                return result
            else:
                return self.getValidOptions(options=options, descriptions=descriptions, filter=result, default=default)
        else:
            if result:
                return result
            else:
                return ""

    def getValidOptions(self, options=[], descriptions=[], bold_descriptions=False, filter="", default="", title="Available Options", text="Select an option:"):
        if not options:
            return ""
        filter = filter.strip().lower()
        if descriptions:
            descriptionslower = [i.lower() for i in descriptions]
            values = [(option, HTML(f"<b>{descriptions[index]}</b>") if bold_descriptions else descriptions[index]) for index, option in enumerate(options) if (filter in option.lower() or filter in descriptionslower[index])]
        else:
            values = [(option, option) for option in options if filter in option.lower()]
        if not values:
            if descriptions:
                values = [(option, HTML(f"<b>{descriptions[index]}</b>") if bold_descriptions else descriptions[index]) for index, option in enumerate(options)]
            else:
                values = [(option, option) for option in options]
        result = radiolist_dialog(
            title=title,
            text=text,
            values=values,
            default=default if default and default in options else values[0][0],
            style=self.style,
        ).run()
        if result:
            notice = f"You've chosen: {result}"
            config.print3(notice)
            return result
        return ""

    def displayFeatureMenu(self, heading, features):
        values = [(command, command if config.terminalDisplayCommandOnMenu else self.parent.dotCommands[command][0]) for command in features]
        result = radiolist_dialog(
            title=heading,
            text="Select a feature:",
            values=values,
            default=features[0],
            style=self.style,
        ).run()
        if result:
            self.parent.printRunningCommand(result)
            return self.parent.getContent(result)
        else:
            return self.parent.exitAction()
            #return ""

    def getMultipleSelection(self, title="Multiple Selection", text="Select item(s):", options=["ALL"], descriptions=[], default_values=["ALL"]) -> list:
        if descriptions:
            values = [(option, descriptions[index]) for index, option in enumerate(options)]
        else:
            values = [(option, option) for option in options]
        return checkboxlist_dialog(
            title=title,
            text=text,
            values=values,
            default_values=default_values,
            style=self.style,
        ).run()



import os, platform

mainFile = os.path.realpath(__file__)
packageFolder = os.path.dirname(mainFile)

# create config.py if it does not exist; in case user delete the file for some reasons
configFile = os.path.join(packageFolder, "config.py")
if not os.path.isfile(configFile):
    open(configFile, "a", encoding="utf-8").close()

try:
    from searchbible import config
except:
    # write off problematic config file
    open(configFile, "w", encoding="utf-8").close()
    from searchbible import config
from searchbible.health_check import HealthCheck

# share mainFile and packageFolder paths in config
config.mainFile = mainFile
config.packageFolder = packageFolder
#package = os.path.basename(config.packageFolder)
if os.getcwd() != config.packageFolder:
    os.chdir(config.packageFolder)

# check current platform
config.thisPlatform = platform.system()
# check if it is running with Android Termux
config.isTermux = True if os.path.isdir("/data/data/com.termux/files/home") else False
# check storage directory
HealthCheck.setSharedItems()
# set up configs
from searchbible.utils.configDefault import *
HealthCheck.check()

# Start of main application
from prompt_toolkit import print_formatted_text, HTML
import chromadb, re, argparse, shutil, threading, asyncio, sys, traceback
from searchbible.chatgpt import ChatGPT
from searchbible.geminipro import GeminiPro
from searchbible.utils.BibleBooks import BibleBooks
from searchbible.utils.BibleVerseParser import BibleVerseParser
from searchbible.utils.prompts import Prompts
from searchbible.utils.prompt_dialogs import TerminalModeDialogs
from searchbible.utils.prompt_validator import NumberValidator, RegexValidator
from searchbible.db.Bible import Bible
from searchbible.utils.AGBsubheadings import agbSubheadings
from searchbible.utils.AGBparagraphs_expanded import agbParagraphs
from searchbible.utils.bible_studies import bible_study_suggestions
from packaging import version
from chromadb.config import Settings
from prompt_toolkit.styles import Style
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.completion import WordCompleter, FuzzyCompleter
from prompt_toolkit.shortcuts import set_title, clear_title
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.input import create_input
from prompt_toolkit.keys import Keys
from pathlib import Path

'''
if not config.isTermux:
    # Work with VLC player for bible audio playback
    import vlc
    # Create a VLC instance
    vlc_instance = vlc.Instance()
    # Create a media player
    media_player = vlc_instance.media_player_new()
'''

appName = "Search Bible AI"
abbrev = BibleBooks.abbrev["eng"]
kjvRefs, _ = BibleBooks().getAllKJVreferences()
config.currentVerses = []

actions = sorted([
    ".configs",
    #".audio",
    ".verses",
    ".paragraphs",
    ".bibles",
    ".chatgpt",
    #".geminipro",
    #".setdefaultchatbot",
])
'''
if not config.isTermux:
    actions.append(".audio")
'''

historyFolder = os.path.join(config.storagedirectory, "history")
Path(historyFolder).mkdir(parents=True, exist_ok=True)
read_history = os.path.join(historyFolder, "read")
read_session = PromptSession(history=FileHistory(read_history))
bible_study_suggestions = [f"{i}[chat]\n" for i in bible_study_suggestions]
config.read_suggestions = Bible.getBibleList() + [i[0] for i in abbrev.values()] + kjvRefs + bible_study_suggestions
read_completer = FuzzyCompleter(WordCompleter(config.read_suggestions + actions, ignore_case=True, sentence=True))
search_book_history = os.path.join(historyFolder, "search_book")
search_book_session = PromptSession(history=FileHistory(search_book_history))
book_suggestions = ["ALL"] + [i[0] for i in abbrev.values()]
book_completer = FuzzyCompleter(WordCompleter(book_suggestions, ignore_case=True))
search_chapter_history = os.path.join(historyFolder, "search_chapter")
search_chapter_session = PromptSession(history=FileHistory(search_chapter_history))
search_literal_history = os.path.join(historyFolder, "search_literal")
search_literal_session = PromptSession(history=FileHistory(search_literal_history))
search_semantic_history = os.path.join(historyFolder, "search_semantic")
search_semantic_session = PromptSession(history=FileHistory(search_semantic_history))
search_closest_match_history = os.path.join(historyFolder, "search_closest_match")
search_closest_match_session = PromptSession(history=FileHistory(search_closest_match_history))
search_regex_history = os.path.join(historyFolder, "search_regex")
search_regex_session = PromptSession(history=FileHistory(search_regex_history))

promptStyle = Style.from_dict({
    # User input (default text).
    "": config.terminalCommandEntryColor2,
    # Prompt.
    "indicator": config.terminalPromptIndicatorColor2,
})

def removeSpecialEntries(content):
    return re.sub("\[chatgpt\]|\[chat\]|\[geminipro\]", "", content)
config.removeSpecialEntries = removeSpecialEntries

def getLastEntry(logFile: str) -> str:
    def isEntry(line):
        line = line.strip()
        return (re.search("^\+[^\.].*?$", line) and not line[1:] in (config.exit_entry, config.cancel_entry))
    if os.path.isfile(logFile) and not os.path.getsize(logFile) == 0:
        with open(logFile, "r", encoding="utf-8") as fileObj:
            lines = tuple(filter(isEntry, fileObj.readlines()))
        if lines:
            return lines[-1][1:].strip()
    return ""

def compareBibles(ref: str, paragraphs: bool=False) -> None:
    if config.compareBibles:
        for i in filter(lambda x: not x == config.mainText, config.compareBibleVersions):
            if data := Bible.getSingleItem(ref, i, paragraphs):
                *_, itemContent = data
                if paragraphs:
                    itemContent = re.sub("\\A.*?\n", "", itemContent.strip())
                    print(f"({i})\n{itemContent}\n")
                else:
                    itemContent = itemContent.strip()
                    config.currentVerses.append((i, ref, itemContent))
                    HealthCheck.print4(f"({i}) {itemContent}")

def read(default: str="") -> None:
    HealthCheck.print2("Search Bible AI")
    HealthCheck.print3("Developed by: Eliran Wong")
    HealthCheck.print3("Open source: https://github.com/eliranwong/searchbibleai")
    HealthCheck.print2(config.divider)
    HealthCheck.print("* enter a single reference to display a full chapter")
    HealthCheck.print("* enter multiple references to display verses")
    HealthCheck.print("* enter a bible version abbreviation, e.g. KJV, to switch version")
    HealthCheck.print("* enter a search query to perform a simple search")
    HealthCheck.print("* enter '.verses' or press 'Ctrl+F' to perform a detailed search for verses")
    HealthCheck.print("* enter '.paragraphs' or press 'Esc+F' to perform a detailed search for paragraphs")
    HealthCheck.print("* press 'Ctrl+K' for more keyboard shortcuts")
    HealthCheck.print(f"* enter '{config.exit_entry}' or press 'Ctrl+Q' to exit current feature of quit this app")
    HealthCheck.print2(config.divider)

    def setDefaultChatbot():
        dialogs = TerminalModeDialogs(None)
        model = dialogs.getValidOptions(
            options=("chatgpt", "geminipro"),
            title="Default Chatbot",
            default=config.chatbot,
            text="Default chatbot is loaded when you include '[chat]' in your input.",
        )
        if model:
            config.chatbot = model
            print(f"Default chatbot: {model}")

    def selectBibleForComparison():
        dialogs = TerminalModeDialogs(None)
        options = Bible.getBibleList()
        versions = dialogs.getMultipleSelection(
            title="Bibles for Comparison",
            text="Select bible version(s):",
            options=options,
            default_values=config.compareBibleVersions,
        )
        if versions:
            config.compareBibleVersions = versions

    if not default:
        default = getLastEntry(read_history)

    parser = BibleVerseParser(config.parserStandarisation)

    this_key_bindings = KeyBindings()
    @this_key_bindings.add("c-f")
    def _(event):
        buffer = event.app.current_buffer
        buffer.text = ".verses"
        buffer.validate_and_handle()
    @this_key_bindings.add("escape", "f")
    def _(event):
        buffer = event.app.current_buffer
        buffer.text = ".paragraphs"
        buffer.validate_and_handle()
    @this_key_bindings.add("c-g")
    def _(event):
        buffer = event.app.current_buffer
        buffer.text = ".chatgpt" if config.chatbot == "chatgpt" else ".geminipro"
        buffer.validate_and_handle()
    @this_key_bindings.add("escape", "g")
    def _(event):
        buffer = event.app.current_buffer
        buffer.text = ".geminipro" if config.chatbot == "chatgpt" else ".chatgpt"
        buffer.validate_and_handle()
    @this_key_bindings.add("c-p")
    def _(event):
        config.compareBibles = not config.compareBibles
        HealthCheck.print3(f"Bible Comparison: {'ON' if config.compareBibles else 'OFF'}")
        buffer = event.app.current_buffer
        buffer.text = getLastEntry(read_history)
        buffer.validate_and_handle()
    @this_key_bindings.add("escape", "p")
    def _(event):
        buffer = event.app.current_buffer
        buffer.text = ".bibles"
        buffer.validate_and_handle()
    @this_key_bindings.add("c-s")
    def _(event):
        config.chapterParagraphsAndSubheadings = not config.chapterParagraphsAndSubheadings
        HealthCheck.print3(f"Chapter Paragraphs and Subheadings: {'ON' if config.chapterParagraphsAndSubheadings else 'OFF'}")
        buffer = event.app.current_buffer
        buffer.text = ":"
        buffer.validate_and_handle()
    if not config.isTermux:
        @this_key_bindings.add("c-y")
        def _(event):
            buffer = event.app.current_buffer
            buffer.text = ".audio"
            buffer.validate_and_handle()
    @this_key_bindings.add("escape", "c")
    def _(event):
        buffer = event.app.current_buffer
        buffer.text = ".configs"
        buffer.validate_and_handle()

    prompts = Prompts(custom_key_bindings=this_key_bindings)

    while True:
        userInput = prompts.simplePrompt(
            style=config.promptStyle1,
            promptSession=read_session,
            completer=read_completer,
            default=default,
            accept_default=True if default else False,
            bottom_toolbar=" [ctrl+q] exit [ctrl+k] shortcut keys ",
        )
        default = ""

        # default chatbot
        userInput = userInput.replace("[chat]", f"[{config.chatbot}]")

        if userInput == config.exit_entry:
            break
        elif userInput == ".configs":
            changeConfigs()
        elif userInput == ".verses":
            # ctrl+f to search verses
            search(bible=config.mainText, paragraphs=False)
        elif userInput == ".paragraphs":
            # ctrl+p to search paragraphs
            search(bible=config.mainText, paragraphs=True)
        elif userInput == ".bibles":
            selectBibleForComparison()
        elif userInput == ".setdefaultchatbot":
            setDefaultChatbot()
        elif userInput == ".chatgpt":
            default = ChatGPT(
                temperature=config.llmTemperature,
                max_output_tokens = config.chatGPTApiMaxTokens,
            ).run()
            """
        elif "[geminipro]" in userInput:
            default = GeminiPro(
                temperature=config.llmTemperature,
                max_output_tokens = config.chatGPTApiMaxTokens,
            ).run(userInput)
        elif userInput == ".geminipro":
            default = GeminiPro(
                temperature=config.llmTemperature,
                max_output_tokens = config.chatGPTApiMaxTokens,
            ).run()
        elif userInput == ".audio" and not config.isTermux:
            playBibleAudio()"""
        elif userInput:
            HealthCheck.print2(config.divider)

            # transform aliases
            bookName = abbrev[str(config.mainB)][0]
            if userInput in Bible.getBibleList() + Bible.getUbaBibleList():
                # change bible version
                config.mainText = userInput
                userInput = f"{bookName} {config.mainC}:{config.mainV}"
            elif userInput == ":":
                # loaded previous selected verse
                userInput = f"{bookName} {config.mainC}:{config.mainV}"
            elif re.search("^[0-9]+?:[0-9]+?$", userInput):
                # change both chapter and verse in the same book
                userInput = f"{bookName} {userInput}"
            elif re.search("^[0-9]+?:$", userInput):
                # change chapter in the same book
                userInput = f"{bookName} {userInput[:-1]}:1"
            elif re.search("^:[0-9]+?$", userInput):
                # change verse in the same book
                userInput = f"{bookName} {config.mainC}:{userInput[1:]}"

            if refs := parser.extractAllReferences(userInput):
                # reset config.currentVerses
                config.currentVerses = []
                # verse reference(s) provided
                isChapter = (len(refs) == 1 and len(refs[0]) == 3)
                if isChapter:
                    # check if it is a single reference; display a full chapter
                    fullChapter, verses = Bible.getVerses(refs, config.mainText)
                    chapterTitle = False
                    for ref, book, chapter, verse, scripture in fullChapter:
                        if not chapterTitle:
                            book_abbr = abbrev[str(book)][0]
                            HealthCheck.print2(f"# {book_abbr} {chapter}")
                            chapterTitle = True
                        elif config.chapterParagraphsAndSubheadings and (book, chapter, verse) in agbParagraphs:
                            print("")
                        if config.chapterParagraphsAndSubheadings and f"{book}.{chapter}.{verse}" in agbSubheadings:
                            HealthCheck.print2(f"## {agbSubheadings[f'{book}.{chapter}.{verse}']}")
                        scripture = scripture.strip()
                        config.currentVerses.append((config.mainText, ref, scripture))
                        HealthCheck.print4(f"({verse}) {scripture}")
                        compareBibles(ref)
                    # draw a whole chapter
                    HealthCheck.print2(config.divider)
                else:
                    verses = Bible.getVerses(refs, config.mainText)
                # display all verses
                book_chapter = ""
                for ref, book, chapter, verse, scripture in verses:
                    this_book_chapter = f"{book}.{chapter}"
                    if not book_chapter:
                        book_chapter = this_book_chapter
                    elif not this_book_chapter == book_chapter:
                        HealthCheck.print2(config.divider)
                        book_chapter = this_book_chapter
                    book_abbr = abbrev[str(book)][0]
                    scripture = scripture.strip()
                    config.currentVerses.append((config.mainText, ref, scripture))
                    HealthCheck.print4(f"({book_abbr} {chapter}:{verse}) {scripture}")
                    compareBibles(ref)
                HealthCheck.print2(config.divider)
            else:
                search(bible=config.mainText, paragraphs=False, simpleSearch=userInput)
                HealthCheck.print2(config.divider)

# change configs
def changeConfigs():
    def loadConfig(configPath):
        with open(configPath, "r", encoding="utf-8") as fileObj:
            configs = fileObj.read()
        configs = "from searchbible import config\n" + re.sub("^([A-Za-z])", r"config.\1", configs, flags=re.M)
        exec(configs, globals())
    # file paths
    configFile = os.path.join(config.packageFolder, 'config.py')
    backupFile = os.path.join(config.storagedirectory, "config_backup.py")
    # backup configs
    HealthCheck.saveConfig()
    shutil.copy(configFile, backupFile)
    # open current configs with built-in text editor
    eTextEditor = f"{sys.executable} {os.path.join(config.packageFolder, 'eTextEdit.py')}"
    os.system(f"{eTextEditor} {configFile}")
    set_title("Search Bible AI")
    # re-load configs
    try:
        loadConfig(configFile)
        HealthCheck.print2("Changes loaded!")
    except:
        HealthCheck.print2("Failed to load your changes!")
        print(traceback.format_exc())
        try:
            HealthCheck.print2("Restoring backup ...")
            loadConfig(backupFile)
            shutil.copy(backupFile, configFile)
            HealthCheck.print2("Restored!")
        except:
            HealthCheck.print2("Failed to restore backup!")

# play bible audio
def playAudioFile(audioFile):
    if not config.isTermux:
        media = vlc_instance.media_new(audioFile)
        media_player.set_media(media)
        media_player.play()
        media_player.set_rate(config.vlcSpeed)
        while config.playback and media_player.get_state() != vlc.State.Ended:
            continue
    else:
        print("This feature is not supported on your device.")

def startBibleAudioPlayback(playback_event):
    for version, ref, scripture in config.currentVerses:
        if config.playback and not playback_event.is_set():
            b, c, v = ref.split(".")
            audioFile = os.path.join(config.storagedirectory, "audio", version, "default", f"{b}_{c}", f"{version}_{b}_{c}_{v}.mp3")
            b = abbrev[str(b)][0]
            if os.path.isfile(audioFile):
                HealthCheck.print2(f"{version} - {b} {c}:{v}")
                HealthCheck.print(scripture)
                playAudioFile(audioFile)
        else:
            break
    if not playback_event.is_set():
        playback_event.set()

def closeMediaPlayer():
    if not config.isTermux:
        media_player.stop()
        config.playback = False
        HealthCheck.print2("\nMedia player stopped!")

def keyToStopPlayback(playback_event):
    # allow users to stop the playback by pressing either ctrl+Q or ctrl+z
    async def readKeys() -> None:
        # create an input
        input = create_input()
        # capture key input
        def keys_ready():
            for key_press in input.read_keys():
                #print(key_press)
                if key_press.key in (Keys.ControlQ, Keys.ControlZ):
                    closeMediaPlayer()
                    playback_event.set()
        # loop when playback is in progress
        with input.raw_mode():
            with input.attach(keys_ready):
                while config.playback and not playback_event.is_set():
                    await asyncio.sleep(0.1)
    # run readKeys
    asyncio.run(readKeys())

def playBibleAudio():
    if not config.isTermux:
        playback_event = threading.Event()
        playback_thread = threading.Thread(target=startBibleAudioPlayback, args=(playback_event,))
        config.playback = True
        playback_thread.start()
        keyToStopPlayback(playback_event)
        playback_thread.join()
    else:
        print("This feature is not supported on your device.")

# combined semantic searches, literal searches and regular expression searches
def search(bible:str="NET", paragraphs:bool=False, simpleSearch="") -> None:

    def getAndItems(query):
        splits = query.split("&&")
        return {"$and": [{"$contains": i} for i in splits]} if len(splits) > 1 else {"$contains": query}

    #dbpath
    dbpath = Bible.getDbPath(bible)
    if not dbpath:
        return None

    # client
    chroma_client = chromadb.PersistentClient(dbpath, Settings(anonymized_telemetry=False))
    # collection
    collection = chroma_client.get_or_create_collection(
        name="paragraphs" if paragraphs else "verses",
        metadata={"hnsw:space": "cosine"},
        embedding_function=HealthCheck.getEmbeddingFunction(embeddingModel="paraphrase-multilingual-mpnet-base-v2"),
    )

    def getChapterRange(i):
        cc = []
        try:
            rangeStart, rangeEnd = [ii.strip() for ii in i.split("-")]
            rangeStart, rangeEnd = int(rangeStart), int(rangeEnd)
            if rangeEnd > rangeStart:
                for ii in range(rangeStart, rangeEnd):
                    cc.append(ii)
                cc.append(rangeEnd)
            else:
                cc.append(rangeStart)
        except:
            pass
        return cc

    if simpleSearch and not paragraphs:
        meaning = simpleSearch.replace("\n", " ")
        n_results = config.maxClosestMatches
        where = None
        contains = None
        regex = None
    else:
        # user input
        HealthCheck.print2(f"SEARCH {'PARAGRAPHS' if paragraphs else 'VERSES'}")
        HealthCheck.print2(config.divider)
        # search in books
        if simpleSearch and paragraphs:
            books = ""
        else:
            HealthCheck.print2("In books (use '||' for combo, '-' for range):")
            print("e.g. Gen||Matt-John||Rev")
            books = HealthCheck.simplePrompt(style=promptStyle, promptSession=search_book_session, completer=book_completer, default=getLastEntry(search_book_history))
            if books.lower() == config.exit_entry:
                return
            if books.lower() == "all":
                books = ""
        books = BibleBooks.getBookCombo(books)
        if books:
            if paragraphs:
                books = {"book_start": {"$in": books}}
            else:
                books = {"book": {"$in": books}}
        else:
            books = {}
        # search in chapters
        if simpleSearch and paragraphs:
            chapters = ""
        else:
            HealthCheck.print2("In chapters (use '||' for combo, '-' for range):")
            print("e.g. 2||4||6-8||10")
            chapters = HealthCheck.simplePrompt(style=promptStyle, promptSession=search_chapter_session, validator=RegexValidator(pattern="^[0-9|]*?$"), default=getLastEntry(search_chapter_history))
            if chapters.lower() == config.exit_entry:
                return
            if chapters.lower() == "all":
                chapters = ""
        if chapters := chapters.strip():
            splits = chapters.split("||")
            if len(splits) == 1:
                if "-" in chapters:
                    cc = getChapterRange(chapters)
                else:
                    try:
                        cc = [int(chapters)]
                    except:
                        cc = []
            else:
                cc = []
                for i in splits:
                    i = i.lower().strip()
                    if "-" in i:
                        cc += getChapterRange(i)
                    else:
                        try:
                            cc.append(int(i))
                        except:
                            pass
        else:
            cc = []
        if cc:
            if paragraphs:
                chapters = {"$or": [{"$and": [{"chapter_start": {"$lte": c}}, {"chapter_end": {"$gte": c}}]} for c in cc]} if len(cc) > 1 else {"$and": [{"chapter_start": {"$lte": cc[0]}}, {"chapter_end": {"$gte": cc[0]}}]}
            else:
                chapters = {"chapter": {"$in": cc}}
        else:
            chapters = {}

        # search for plain words
        if simpleSearch and paragraphs:
            contains = ""
        else:
            HealthCheck.print2("Search for plain words ('||' denotes 'or'; '&amp;&amp;' denotes 'and'):")
            print("e.g. Lord&&God||Jesus&&love")
            contains = HealthCheck.simplePrompt(style=promptStyle, promptSession=search_literal_session, default=getLastEntry(search_literal_history))
            if contains.lower() == config.exit_entry:
                return
            if contains.strip():
                splits = contains.split("||")
                contains = {"$or": [getAndItems(i) for i in splits]} if len(splits) > 1 else getAndItems(contains)
            else:
                contains = ""
        # search for meaning
        if simpleSearch and paragraphs:
            meaning = simpleSearch
        else:
            HealthCheck.print2("Search for meaning:")
            meaning = HealthCheck.simplePrompt(style=promptStyle, promptSession=search_semantic_session, default=getLastEntry(search_semantic_history))
            if meaning.lower() == config.exit_entry:
                return
        if meaning:
            if not (simpleSearch and paragraphs):
                HealthCheck.print2("Maximum number of closest matches:")
                # specify number of closest matches
                default_n_results = config.maxClosestMatches
                n_results = HealthCheck.simplePrompt(style=promptStyle, promptSession=search_closest_match_session, validator=NumberValidator(), default=str(default_n_results))
                if n_results.lower() == config.exit_entry:
                    return
                if n_results and int(n_results) > 0:
                    config.maxClosestMatches = int(n_results)
                else:
                    config.maxClosestMatches = default_n_results
        # search for regex
        if simpleSearch and paragraphs:
            regex = ""
        else:
            HealthCheck.print2("Search for regular expression:")
            regex = HealthCheck.simplePrompt(style=promptStyle, promptSession=search_regex_session, default=getLastEntry(search_regex_history))
            if regex.lower() == config.exit_entry:
                return

        # formulate where filter
        if books and chapters:
            where = {"$and": [books, chapters]}
        elif books:
            where = books
        elif chapters:
            where = chapters
        else:
            where = None

    if meaning:
        # run query
        res = collection.query(
            query_texts=[meaning],
            n_results = config.maxClosestMatches,
            where=where,
            where_document=contains if contains else None,
        )
    else:
        res = collection.get(
            where=where,
            where_document=contains if contains else None,
        )
    if not simpleSearch:
        HealthCheck.print2(config.divider)
    HealthCheck.print2(f">>> Retrieved {'paragraphs' if paragraphs else 'verses'}:\n")

    if meaning:
        metadatas = res["metadatas"][0]
        documents = res["documents"][0]
    else:
        metadatas = res["metadatas"]
        documents = res["documents"]

    if paragraphs:
        verses = [(metadata["start"], metadata["book_start"], metadata["chapter_start"], metadata["verse_start"], metadata["chapter_end"], metadata["verse_end"], document) for metadata, document in zip(metadatas, documents)]
    else:
        verses = [(metadata["reference"], metadata["book"], metadata["chapter"], metadata["verse"], document) for metadata, document in zip(metadatas, documents)]

    plainTextOutput = []

    if not meaning:
        # sorting for non-semantic search
        verses = sorted(verses, key=lambda x: version.parse(x[0]))

    if paragraphs:
        for ref, book, chapter, verse, chapter_end, verse_end, scripture in verses:
            book_abbr = abbrev[str(book)][0]
            if not regex or (regex and re.search(regex, scripture, flags=re.I|re.M)):
                title = f"# {book_abbr} {chapter}:{verse}-{chapter_end}:{verse_end}"
                plainTextOutput += [title, scripture]
                HealthCheck.print2(title)
                scripture = re.sub(r"\A(.+?)$", r"<{0}>## \1</{0}>".format(config.terminalPromptIndicatorColor2), scripture, flags=re.M)
                scripture = re.sub("^([0-9]+?:[0-9]+?) ", r"<{0}>(\1)</{0}>".format(config.terminalPromptIndicatorColor2), scripture, flags=re.M)
                print_formatted_text(HTML(f"{scripture.strip()}\n"))
                compareBibles(ref, paragraphs=True)
    else:
        for ref, book, chapter, verse, scripture in verses:
            book_abbr = abbrev[str(book)][0]
            if not regex or (regex and re.search(regex, scripture, flags=re.IGNORECASE)):
                scripture = scripture.strip()
                config.currentVerses.append((config.mainText, ref, scripture))
                content = f"({book_abbr} {chapter}:{verse}) {scripture}"
                HealthCheck.print4(content)
                plainTextOutput.append(content)
                compareBibles(ref)
    
    if not simpleSearch:
        HealthCheck.print2(config.divider)
    return "\n".join(plainTextOutput)

def main():
    # Create the parser
    parser = argparse.ArgumentParser(description="SearchBibleAI CLI options")
    # Add arguments
    parser.add_argument("default", nargs="?", default=None, help="Specify a bible module, e.g. KJV, NET, etc.")
    # Parse arguments
    args = parser.parse_args()
    # Get options
    default = args.default.strip() if args.default and args.default.strip() else ""

    # set terminal title
    set_title("Search Bible AI")
    # check log files; delete old lines if too many
    for i in (
        read_history,
        search_book_history,
        search_chapter_history,
        search_literal_history,
        search_semantic_history,
        search_closest_match_history,
        search_regex_history,
    ):
        HealthCheck.set_log_file_max_lines(i, 3000)
    # start with reading mode
    read(default=default)
    # back up config on closing
    HealthCheck.print2("Saving configurations ...")
    HealthCheck.saveConfig()
    shutil.copy(configFile, os.path.join(config.storagedirectory, "config_backup.py"))
    # clear terminal title
    HealthCheck.print2("Closing ...")
    clear_title()

    if not config.isTermux:
        # Release the media player
        media_player.release()
        # Release the VLC instance
        vlc_instance.release()


if __name__ == '__main__':
    main()